<?php

App::uses('AppHelper', 'View/Helper');

class ImgHelper extends AppHelper {

	public $helpers = array(
		'Html'
	);

	public function __construct(View $View, $settings = []) {
		if ($settings == null) {
			$settings = array();
		}

		parent::__construct($View, $settings);

		$defaults = array(
			'uploadDirectory' => 'files' . DS . 'image',
			'remoteCacheDirectory' => 'files' . DS . 'remote_cache',
			'cacheDirectory' => 'files' . DS . 'img_cache',
			'memoryLimit' => '1024M',
			'quality' => 75,
			'fullBase' => false
		);

		$this->settings = array_merge($defaults, $settings);

		return;
	}

	/**
	 * Resizes an image.
	 *
	 * @param array|string Image to resize
	 * @param array $dimensions Dimensions for resized image
	 * @param array $options Img tag attributes
	 * @param string $context Image context (defines a directory to place image)
	 * @return string fully qualified img tag
	 */
	public function resize($image, $dimensions = null, $options = [], $context = null) {
		$fullBase = isset($options['fullBase']) ? $options['fullBase'] : $this->settings['fullBase'];
		$path = $this->path($image, $dimensions, $context, $fullBase);

		unset($options['fullBase']);

		if (!empty($image['alt']) && empty($options['alt'])) {
			$options['alt'] = $image['alt'];
		} elseif (empty($options['alt']) && !empty($image['name'])) {
			$options['alt'] = $image['name'];
		}

		// Make sure we apply the dimension attributes to the output image to
		// improve browser performance (this allows the browser to know how
		// much space to allocate the image before it fully renders).
		if (empty($options['width']) && !empty($dimensions['width'])) {
			$options['width'] = $dimensions['width'];
		}
		if (empty($options['height']) && !empty($dimensions['height'])) {
			$options['height'] = $dimensions['height'];
		}

		// Immediate can be used to skip any lazy loading set up
		if (Configure::read('EvImg.lazyload') && empty($options['immediate'])) {
			$options['data-src'] = $path;
			$options['data-lazy'] = true;
			$path = !empty($options['placeholder']) ?
				$options['placeholder'] :
				"data:image/svg+xml,%3Csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 0 11 14'%3E%3C/svg%3E";

			$this->Html->script('EvImg.lazyload.js', ['block' => 'script']);
		}

		unset($options['immediate']);
		unset($options['placeholder']);

		return $this->Html->image($path, $options);
	}

	/**
	 * Wrapper method so that we can directly get a path to a cached resized image. This is to get
	 * round issues with the second parameter needing to be passed by reference.
	 *
	 * @param array|string Image to resize
	 * @param array $dimensions Dimensions for resized image
	 * @param string $context Image context (defines a directory to place image)
	 * @return string Path to resized image
	 */
	public function path($image, $dimensions = null, $context = null, $fullBase = false) {
		$path = $this->_path($image, $dimensions, $context);
		return $this->Html->assetUrl($path, compact('fullBase'));
	}

	/**
	 * Returns the path to a cached resized image.
	 *
	 * @param array|string Image to resize
	 * @param array $dimensions Dimensions for resized image
	 * @param string $context Image context (defines a directory to place image)
	 * @return string Path to resized image
	 */
	protected function _path($image, &$dimensions = null, $context = null) {
		$isPdf = !empty($dimensions['pdf']);

		$cacheFilepath = $this->settings['cacheDirectory'] . DS;

		if (! empty($image)) {
			if (is_array($image)) {

				if (strpos($image['filename'], '://') !== false) {
					$originalFilepath = $this->_cacheRemoteImage($image['filename']);
					$originalFilename = basename($originalFilepath);
					$originalFilepath = str_replace($originalFilename, '', $originalFilepath);
				} else {
					$originalFilename = $image['filename'];
					$originalFilepath = $this->settings['uploadDirectory'] . DS . $image['id'] . DS;
				}

				$cacheFilepath .= ($context !== null ? $context : $image['id']) . DS;

				$isPdf = in_array($image['type'], ['application/pdf', 'application/postscript']);

			} elseif (strpos($image, '://') !== false) {

				$originalFilepath = $this->_cacheRemoteImage($image);
				$originalFilename = basename($originalFilepath);
				$originalFilepath = str_replace($originalFilename, '', $originalFilepath);
				$cacheFilepath .= $context !== null ? $context . DS : '';

			} else {

				$parts = pathinfo($image);
				$originalFilename = $parts['basename'];
				$originalFilepath = DS . $parts['dirname'] . DS;
				$cacheFilepath .= $context !== null ? $context . DS : '';

			}
		}

		// If no image exists, setup no image placeholder
		if (
			empty($image) ||
			! file_exists(WWW_ROOT . $originalFilepath . $originalFilename)
		) {
			$originalFilepath = 'img' . DS;

			$originalFilename = 'no-image.png';

			// check whether we have a themed version available
			if (! empty($this->theme)) {
				$themeDir = 'theme' . DS . $this->theme;

				// check whether the no-image.png file exists for the theme,
				// replacing the $originalFilepath when available
				if (file_exists($themeDir . DS . $originalFilepath . $originalFilename)) {
					$originalFilepath = $themeDir . DS . $originalFilepath;
				}
			}
		}

		// Create a filename for the cached image.
		$cacheOptions = array_intersect_key(
			$dimensions,
			array_flip(array('width', 'height', 'crop', 'fit', 'blur'))
		);

		// Flatten options into a single-level array so we can implode them all
		$cacheOptionsFlattened = [];
		array_walk_recursive($cacheOptions, function ($v) use (&$cacheOptionsFlattened) {
			$cacheOptionsFlattened[] = $v;
		});

		$cacheFilename = $cacheFilepath . implode('_', $cacheOptionsFlattened) . '_' . $originalFilename;

		$convertToPng = $this->_shouldConvertToPng(WWW_ROOT . $originalFilepath . $originalFilename);
		if ($convertToPng) {
			$cacheFilename .= '.png';
		}

		if (!file_exists(WWW_ROOT . $cacheFilename)) {

			$Image = new imagick();
			try {
				if ($isPdf === true) {
					$Image->setResolution(300, 300);
					$Image->setBackgroundColor('#ffffff');
					$Image->readImage(WWW_ROOT . $originalFilepath . $originalFilename . '[0]');
					$Image = $Image->mergeImageLayers(Imagick::LAYERMETHOD_FLATTEN);
				} else {
					$Image->readImage(WWW_ROOT . $originalFilepath . $originalFilename);
				}
			} catch (ImagickException $e) {
				CakeLog::write('error', 'EvImg: Invalid Image Format for file ' . $originalFilepath . $originalFilename);

				// Show placeholder image instead.
				$Image->readImage(WWW_ROOT . DS . 'img' . DS . 'no-image.png');
			}

			$height = $Image->getImageHeight();
			$width = $Image->getImageWidth();

			if (!empty($dimensions['background'])) {
				$Image->setbackgroundcolor($dimensions['background']);
				$Image->setimagebackgroundcolor($dimensions['background']);
			}

			if (!empty($dimensions['crop'])) {

				if (!empty($dimensions['cropPosition'])) {
					$this->_cropFromPosition($Image, $dimensions, $width, $height);
				} else {
					// default to centre if no crop position is specified
					$Image->cropThumbnailImage($dimensions['width'], $dimensions['height']);
				}

			} elseif (!empty($dimensions['fill'])) {

				// Fully resize to given width and/or height even if original image is smaller
				$Image->thumbnailImage(
					!empty($dimensions['width']) ? (int)$dimensions['width'] : 0,
					!empty($dimensions['height']) ? (int)$dimensions['height'] : 0,
					false,
					true
				);

			} elseif (empty($dimensions['width'])) {

				if ((int)$dimensions['height'] < $height) {
					$Image->thumbnailImage(0, (int)$dimensions['height']);
				}

			} elseif (empty($dimensions['height'])) {

				if ((int)$dimensions['width'] < $width) {
					$Image->thumbnailImage((int)$dimensions['width'], 0);
				}

			} elseif (!empty($dimensions['fit'])) {
				// Check if the original image is smaller than the desired size. If it is, pad around the image with a
				// border.
				if ($width <= $dimensions['width'] && $height < $dimensions['height']) {

					$background = (isset($dimensions['background']) ? $dimensions['background'] : 'white');

					$Image->scaleImage($dimensions['width'], $dimensions['height'], true);

					$borderWidth = max(($dimensions['width'] - $width) / 2, 0);
					$borderHeight = max(($dimensions['height'] - $height) / 2, 0);

					$Image->borderImage($background, $borderWidth, $borderHeight);
				} else {
					// Resize image to given dimensions but doesn't stretch larger than original
					$Image->thumbnailImage($dimensions['width'], $dimensions['height'], true, true);
				}
			} elseif ($height <= $width) {

				if ((int)$dimensions['width'] < $width) {
					$Image->thumbnailImage((int)$dimensions['width'], 0);
				}

			} else {

				if ((int)$dimensions['height'] < $height) {
					$Image->thumbnailImage(0, (int)$dimensions['height']);
				}

			}

			if (!empty($dimensions['blur'])) {
				$radius = isset($dimensions['blur']['radius']) ? $dimensions['blur']['radius'] : 5;
				$sigma = isset($dimensions['blur']['sigma']) ? $dimensions['blur']['sigma'] : 3;
				$Image->blurImage($radius, $sigma);
			}

			if (!file_exists(WWW_ROOT . $cacheFilepath)) {
				// cache directory doesn't exist - create it now
				mkdir(WWW_ROOT . $cacheFilepath, 0777, true);
			}

			$dimensions = $Image->getImageGeometry();

			$Image->setImageCompressionQuality($this->settings['quality']);

			// Strip image of all profiles and comments to reduce size.
			$Image->stripImage();

			// Convert non-websafe formats to png. This is anything other than png, jpg, svg and gif.
			if ($convertToPng) {
				$Image->setImageFormat('png');
			}

			$Image->writeImage(WWW_ROOT . $cacheFilename);

		} else {
			// Correct image dimensions to be used on the <img/> tag.
			$Image = new imagick();
			$Image->readImage(WWW_ROOT . $cacheFilename);
			$dimensions = $Image->getImageGeometry();
		}

		$cdnAssetsUrl = Configure::read('cdn_assets_url');
		if (! empty($cdnAssetsUrl)) {
			//If a cdn asset url is available, use that instead of the base url.
			$cacheFilename = $cdnAssetsUrl . '/' . $cacheFilename;
		} else {
			$cacheFilename = '/' . $cacheFilename;
		}

		// Return the image path (make sure we use forward and not back slashes
		// for the image, this is an issue when using a Windows server).
		return str_replace('\\', '/', $cacheFilename);
	}

	public function getDimensions($image) {
		$Image = new imagick();
		$Image->readImage($image['filepath']);
		$dimensions = $Image->getImageGeometry();

		return $dimensions;
	}

/**
 * Determines if a given file will be converted to a PNG.
 *
 * @param string $filepath The filepath to the image to check
 * @return boolean true for convert, false to use the current format
 */
	protected function _shouldConvertToPng($filepath) {
		// Check the mime type. If it is not a jpg, gif svg or png convert to png.
		// For mimetypes see here: https://www.sitepoint.com/mime-types-complete-list/
		switch (mime_content_type( $filepath )) {
			case 'image/png':
			case 'image/jpeg':
			case 'image/gif':
			case 'image/svg+xml':
				// Do nothing. These formats will be preserved;
				return false;
			default:
				// Any other format should be converted to png
				return true;
		}
	}

/**
 * Resizes image to user defined dimensions, either by width or height depending on which is greater
 * @param  $Image - Imagick instance
 * @param  $dimensions - user defined dimensions
 * @param  $width - image width
 * @param  $height - image height
 * @return $Image - modified Imagick instance
 */
	protected function _scaleDown($Image, $dimensions, $width, $height) {
		if ((int)$dimensions['width'] < $width) {
			$Image->thumbnailImage((int)$dimensions['width'], 0);
		} elseif ((int)$dimensions['height'] < $height) {
			$Image->thumbnailImage(0, (int)$dimensions['height']);
		}

		return $Image;
	}

/**
 * Handles cropping an image from a specified position
 * The default crop method used doesn't allow for dimensions to be passed through so we break it up into 2 Imagick methods
 * Imagick::thumbnailImage & Imagick::cropImage
 * @param  $Image - Imagick instance
 * @param  $dimensions - user defined dimensions
 * @param  $width - image width
 * @param  $height - image height
 * @return $Image - modified Imagick instance
 */
	protected function _cropFromPosition($Image, $dimensions, $width, $height) {
		// cropPosition can either be a string value (top|middle|bottom) - or it can be an array with x and y keys
		if (is_array($dimensions['cropPosition'])) {
			// default x and y to 0 if neither are passed
			$x = !empty($dimensions['cropPosition']['x']) ? $dimensions['cropPosition']['x'] : 0;
			$y = !empty($dimensions['cropPosition']['y']) ? $dimensions['cropPosition']['y'] : 0;

			$this->_scaleDown($Image, $dimensions, $width, $height);
			$Image->cropImage($dimensions['width'], $dimensions['height'], $x, $y);
		} else {
			// crop from particular position according to passed value
			switch($dimensions['cropPosition']) {
				case 'top':
					$this->_scaleDown($Image, $dimensions, $width, $height);
					$Image->cropImage($dimensions['width'], $dimensions['height'], 0, 0);
					break;
				case 'bottom':
					$this->_scaleDown($Image, $dimensions, $width, $height);
					// Get the height after resize
					$height = $Image->getImageHeight();
					// 0, 0 is top left. So in order to crop from the bottom we set the y co-ordinate as (image height - expected height)
					$Image->cropImage($dimensions['width'], $dimensions['height'], 0, $height - $dimensions['height']);
					break;
				case 'middle':
				default:
					$Image->cropThumbnailImage($dimensions['width'], $dimensions['height']);
					break;
			}
		}

		return $Image;
	}

/**
 * Locally cache a remote image so that we can resize it.
 * @param string $file File path
 * @return string Locally hosted file path
 */
	protected function _cacheRemoteImage($file) {
		if (! file_exists(WWW_ROOT . $this->settings['remoteCacheDirectory'])) {
			mkdir(WWW_ROOT . $this->settings['remoteCacheDirectory'], 0777, true);
		}
		try {
			$filepath = $this->settings['remoteCacheDirectory'] . DS . md5($file);
			if (! file_exists($filepath)) {
				// Surpress errors when copying, we will catch an exception and return the
				// placeholder image if the file cannot be retrieved.
				@copy($file, $filepath);
			}
		} catch (Exception $e) {
			$filepath = 'img' . DS . 'no-image.png';
		}

		return $filepath;
	}

}
